/*
    ESP32 Antique Radio

    Built into a laser cut antique radio case
    
    This program brings together the functionality of
    the Internet Radio, the SD and the Network/Web
    Music players. Once in a mode you must double
    click the Volume button to change (reboot).

    The VS1053 can play the following formats:
      Ogg Vorbis/MP3/AAC/WMA/FLAC/WAV among others

    Antique Radio hardware consists of:
      1. ESP32 VROOM Module
      2. MCUDEV_TFT1.44 LCD SPI display 128x128
      3. Adafruit VS1053 module #1381
      4. 16GB SD memory card
      5. 2 rotary encoders with switches
      6. Stereo output jack

    See Hardware.h for connections between ESP32 and 
    the display module, the VS1053 module and the 
    rotary encoders.

    Compile using the uPesy ESP32 Vroom DevKit board type
    Partition Scheme: Huge App

    Concept, Design and Implementation: Craig A. Lindley
    Last Update: 12/15/2023
*/

#include <PinButton.h>
#include <SD.h>
#include <WiFiManager.h>  // https://github.com/tzapu/WiFiManager

// Include code common for all music player varieties
#include "Hardware.h"
#include "RotaryEncoder.h"
#include "Navigation.h"
#include "DisplayHelper.h"
#include "ListBox.h"
#include "VS1053.h"

#ifndef min
#define min(a, b) (((a) < (b)) ? (a) : (b))
#endif

// Version Numbers
#define MAJOR_VERSION 0
#define MINOR_VERSION 9
#define MICRO_VERSION 0

// Timeout for the LCD display
#define DISPLAY_TIMEOUT_MIN 1
#define DISPLAY_TIMEOUT_MS (DISPLAY_TIMEOUT_MIN * 60 * 1000)

// Duration of info screen display
#define INFO_SCREEN_DELAY_MS 2000

// Line height on display
#define LINE_HEIGHT 10

// Number of usable lines in list box in addition to title
#define LB_LINES 11

// Max number of chars/line in list box
#define LB_LINE_LENGTH 18

// Forward definition of ListBox painter
void paintListBox(boolean, uint32_t);

/************************************************************************/
/*                       Instantiate Driver Objects                     */
/************************************************************************/

// TFT Display Driver for ST7735 controller on LCD display
DisplayHelper tft(TFT_CS, TFT_DC, TFT_RST, TFT_BL);

// Instantiate listbox
ListBox listBox(LB_LINES, paintListBox);

// Instantiate drivers for the rotary encoders
RotaryEncoder volRE(VOL_RE_PLUS, VOL_RE_MINUS);
RotaryEncoder songsRE(SONGS_RE_PLUS, SONGS_RE_MINUS);

// Instantiate drivers for rotary encoder switches as PinButtons
PinButton volSW(VOL_RE_SW);
PinButton songsSW(SONGS_RE_SW);

// Instantiate navigation instance for each rotary encoder
// Navigation volNav(&volSW, &volRE);
Navigation songsNav(&songsSW, &songsRE);

// Instantiate VS1053 music module
VS1053 vs;

// Include code specific to each music player
#include "MiscCode.h"
#include "IR_Player.h"
#include "SD_Player.h"
#include "WEB_Player.h"

// Instantiate player objects
IRPlayer irPlayer(&vs);
SDPlayer sdPlayer(&vs);
WEBPlayer webPlayer(&vs);

/************************************************************************/
/*                         Constants and Variables                      */
/************************************************************************/

// Task to move audio data
TaskHandle_t Task0;

// Three possible modes for Antique Radio operation
enum PLAYER_MODES {
  IR_MODE,  // Internet Radio mode
  SD_MODE,  // Play music from SD card mode
  WB_MODE   // Go to craigandheather.net and play my tunes mode
};

// Current operating mode
enum PLAYER_MODES playerMode;

/************************************************************************/
/*                               Misc Functions                         */
/************************************************************************/

// Display WiFi connect screen
void displayWiFiScreen() {

  // Clear the display
  tft.clearScreen();
  tft.drawRect(0, 0, tft.width(), tft.height(), COLOR_BLUE);

  // Draw string
  tft.drawCenteredText(calcLineOffset(2), "- WiFi -");

  // Draw message
  tft.drawCenteredText(calcLineOffset(4), "Connecting");
}

// Display welcome screen
void displayWelcomeScreen(void) {

  char buffer[20];
  sprintf(buffer, "Version: %d.%d.%d", MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION);

  // First clear the screen
  tft.clearScreen();

  // Draw border
  tft.drawRect(0, 0, tft.width(), tft.height(), COLOR_BLUE);

  tft.drawCenteredText(calcLineOffset(1), "A Modern");
  tft.drawCenteredText(calcLineOffset(2), "Antique Radio ");
  tft.drawCenteredText(calcLineOffset(6), buffer);
  tft.drawCenteredText(calcLineOffset(8), "Craig A. Lindley");
}

// Display the nav screen
void displayNavScreen(void) {

  // First clear the screen
  tft.clearScreen();
  tft.drawRect(0, 0, tft.width(), tft.height(), COLOR_BLUE);

  // Draw titles
  tft.drawCenteredText(calcLineOffset(1), "Song Control");
  tft.drawCenteredText(calcLineOffset(2), "Rotary Encoder");
  tft.drawCenteredText(calcLineOffset(3), "- Navigation -");

  // Label the actions
  tft.drawCenteredText(calcLineOffset(5), "Forward: RE Right");
  tft.drawCenteredText(calcLineOffset(7), "Backward: RE Left");
  tft.drawCenteredText(calcLineOffset(9), "Select: RE click");
  tft.drawCenteredText(calcLineOffset(11), "Back: RE dbl click");
}

// Display mode selection screen and get selection
int getModeSelection() {

  // Display the mode menu ListBox
  listBox.setDataSource(OPERATIONS_DS);
  listBox.clear();
  listBox.setCenterFlag(true);
  listBox.setTitle("- Mode -");

  // Populate list box with possible modes
  listBox.addEntry("Internet Radio");
  listBox.addEntry("SD Player");
  listBox.addEntry("Web Player");

  listBox.doRepaint(true);

  bool done = false;

  while (!done) {

    // Update state of the rotary encoder switches
    songsSW.update();

    // Poll the songs navigational rotary encoder
    byte result = songsNav.getNavigationDirection();

    if (result == Navigation::NAV_UP) {
      listBox.selectionDown(true);
    }

    else if (result == Navigation::NAV_DOWN) {
      listBox.selectionUp(true);
    }

    else if (result == Navigation::NAV_BACK) {
      // Nothing to do here
    }

    else if (result == Navigation::NAV_SELECT) {
      done = true;
    }
    delay(1);
  }
  return listBox.getSelectionIndex();
}

/************************************************************************/
/*                             Program Setup                            */
/************************************************************************/

void setup() {

  // Setup serial communication for debugging
  Serial.begin(115200);
  delay(1000);
  Serial.println("\nAntique Radio");

  // Initialize the random number generator
  randomSeed(time(NULL));

  // Initialize the SPI interface
  SPI.end();
  SPI.begin(SPI_CLK, SPI_MISO, SPI_MOSI);
  SPI.setHwCs(false);

  // Initialise the SD card interface. Retry as necessary
  int trys = 10;
  while (trys--) {
    if (SD.begin(VS_SDCS)) {
      break;
    }
    delay(500);
  }
  if (trys == 0) {
    Serial.println("SD Card Mount Failed");
    while (true) {};
  }
  Serial.println("SD Card Mounted Successfully\n");

  // Initialize ST7735 display controller
  tft.begin();

  // Indicate we are trying to establish a WiFi connection
  displayWiFiScreen();

  // WiFiManager local intialization
  // Once its business is done, there is no need to keep it around
  WiFiManager wm;

  // Reset settings - wipe stored credentials for testing
  // these are stored by the esp library
  // wm.resetSettings();

  // Automatically connect using saved credentials
  // If connection fails, it starts an access point with the specified
  // name "AntiqueRadio" and then goes into a blocking loop awaiting
  // configuration and will return success result
  if (!wm.autoConnect("AntiqueRadio")) {
    Serial.println("WiFi Connection failed");
    delay(1000);
    ESP.restart();
  } else {
    // If we get here we are golden
    Serial.println("WiFi Connection Established\n");
  }

  // Display initial screens
  displayWelcomeScreen();
  delay(INFO_SCREEN_DELAY_MS);

  displayNavScreen();
  delay(INFO_SCREEN_DELAY_MS);

  // Start the VS1053 module
  vs.begin();

  // Output a test tone to verify operation
  // vs.sineTest(3, 250);

  // Set initial volume in volume rotary encoder
  volRE.setEncoderValue(DEFAULT_VOLUME);

  // Get mode of operation
  switch (getModeSelection()) {
    case 0:
      playerMode = IR_MODE;
      irPlayer.irSetup();
      break;

    case 1:
      playerMode = SD_MODE;
      sdPlayer.sdSetup();
      break;

    case 2:
      playerMode = WB_MODE;
      webPlayer.webSetup();
      break;
  }

  // Create a separate task to move audio data
  // The task will execute the Task0code() function, with priority 0 and executed on core 0
  xTaskCreatePinnedToCore(
    Task0code, /* Task function. */
    "Task0",   /* name of task. */
    10000,     /* Stack size of task */
    NULL,      /* parameter of the task */
    0,         /* priority of the task */
    &Task0,    /* Task handle to keep track of created task */
    0);        /* pin task to core 0 */

  delay(500);
}

// Task0code running on Core 0
void Task0code(void *pvParameters) {
  //  Serial.print("Task0 running on core ");
  //  Serial.println(xPortGetCoreID());

  for (;;) {
    // If playing, move the appropriate audio data to the VS1053 module
    if (vs.isPlaying()) {
      switch (playerMode) {
        case IR_MODE:
          irPlayer.moveData();
          break;
        case SD_MODE:
          sdPlayer.moveData();
          break;
        case WB_MODE:
          webPlayer.moveData();
          break;
      }
    }
    delay(1);
  }
}

/************************************************************************/
/*                             Program Loop                             */
/************************************************************************/

void loop() {

  // Update state of the rotary encoder switches
  songsSW.update();
  volSW.update();

  // Check for and process volume change
  if (volRE.readEncoder(4, 0, 100)) {
    vs.setVolume(volRE.getEncoderValue());
  }

  // If volume rotary encoder double clicked, restart
  if (volSW.isDoubleClick()) {
    ESP.restart();
  }

  // Dispatch to selected player
  switch (playerMode) {
    case IR_MODE:
      irPlayer.irLoop();
      break;
    case SD_MODE:
      sdPlayer.sdLoop();
      break;
    case WB_MODE:
      webPlayer.webLoop();
      break;
  }
  delay(10);
}
