/*
   ESP32 Antique Radio Player - Display Helper Class

   NOTE: with textSize == 1 display can display 13 lines of 21 chars max
   with 17 pixels between lines.

   Concept, Design and Implementation: Craig A. Lindley
   Last Update: 11/21/2023
*/

#ifndef DISPLAYHELPER_H
#define DISPLAYHELPER_H

#include <Adafruit_GFX.h>     // Core graphics library
#include <Adafruit_ST7735.h>  // Hardware-specific library for ST7735


// Some ready-made 16-bit ('565') colors
#define COLOR_BLACK 0x0000
#define COLOR_WHITE 0xFFFF
#define COLOR_RED 0xF800
#define COLOR_GREEN 0x07E0
#define COLOR_BLUE 0x001F
#define COLOR_CYAN 0x07FF
#define COLOR_MAGENTA 0xF81F
#define COLOR_YELLOW 0xFFE0
#define COLOR_ORANGE 0xFC00

// Display Helper Class
class DisplayHelper : public Adafruit_ST7735 {

public:

  // Class Constructor
  DisplayHelper(uint8_t cs, uint8_t dc, uint8_t mosi, uint8_t clk,
                uint8_t rst, uint8_t bl)
    : Adafruit_ST7735(cs, dc, mosi, clk, rst) {
    // Save incoming
    _bl = bl;

    // Configure backlight GPIO pin as output
    pinMode(_bl, OUTPUT);
  }

  DisplayHelper(uint8_t cs, uint8_t dc, uint8_t rst, uint8_t bl)
    : Adafruit_ST7735(cs, dc, rst) {
    // Save incoming
    _bl = bl;

    // Configure backlight GPIO pin as output
    pinMode(_bl, OUTPUT);
  }

  // Initialize the tft display
  void begin() {
    // Initialize ST7735 into 128x128 mode
    initR(INITR_144GREENTAB);

    // Rotate display so USB connector is on the right
    setRotation(3);

    clearScreen();

    setTextSize(1);
    setTextColor(COLOR_WHITE);

    // Configure backlight GPIO pin as output
    pinMode(_bl, OUTPUT);

    // Turn on backlight
    setBacklight(HIGH);
  }

  void clearScreen() {
    fillScreen(COLOR_BLACK);
  }

  // Set the state of the backlight
  // HIGH is on; LOW is off
  void setBacklight(boolean state) {
    digitalWrite(_bl, state);
  }

  // Get the width in pixels of a text string in current font
  uint16_t getTextWidth(const char *text) {

    int16_t x1, y1;
    uint16_t w, h;

    getTextBounds(text, 0, 0, &x1, &y1, &w, &h);

    return w;
  }

  // Get the height in pixels of a text string in current font
  uint16_t getTextHeight(const char *text) {

    int16_t x1, y1;
    uint16_t w, h;

    getTextBounds(text, 0, 0, &x1, &y1, &w, &h);

    return h;
  }

  // Calculate the x offset necessary to center a text string
  int xOffsetToCenterText(const char *text) {
    return (width() - getTextWidth(text)) / 2;
  }

  // Draw a text string at specified location
  void drawText(int x, int y, const char *text) {
    setCursor(x, y);
    print(text);
  }

  // Draw a text string centered on display with y position
  void drawCenteredText(int y, const char *text) {
    drawText(xOffsetToCenterText(text), y, text);
  }

protected:
  uint8_t _bl;
};

#endif
