/*
 * This class encapsulates the Internet Radio Functionality
 *
 * Concept, Design and Implementation by: Craig A. Lindley
 * Last Update: 12/15/2023
 */

#ifndef IR_PLAYER
#define IR_PLAYER

#include "URLParser.h"
#include "StationGatherer.h"

class IRPlayer {

public:
  // Class constructor
  IRPlayer(VS1053 *_pVS1053) {
    pVS1053 = _pVS1053;
  }

  // Determine if a song is still playing
  bool isPlaying() {
    return pVS1053->isPlaying();
  }

  bool startPlayback(const char *url) {

    // Get the current volume level
    int vol = pVS1053->getVolume();

    // Lower volume during change of stream
    pVS1053->setVolume(0);

    // Stop playback
    pVS1053->stopPlayback();

    if (client.connected()) {
      // Disconnect from current connection
      client.stop();
      delay(50);
    }

    // Clear audio data buffer
    memset(buffer, 0, sizeof(buffer));

    // Parse URL into component parts
    parseURL(url, &urlStruct);

    /*
    Serial.printf("Protocol: %s, Host: %s, Port: %s, Path: %s\n",
                  urlStruct.protocol.c_str(),
                  urlStruct.host.c_str(),
                  urlStruct.port.c_str(),
                  urlStruct.path.c_str());
    */
    int port = 80;

    if (urlStruct.port != "") {
      port = atoi(urlStruct.port.c_str());
    }

    if (!client.connect(urlStruct.host.c_str(), port)) {
      Serial.println("Connection failed");
      return false;
    }
    // Serial.println("Connection to host successful\n");

    delay(50);

    // Form the GET string. Host must be specified or
    // server won't recognize the request
    String getString = String("GET ");
    getString += urlStruct.path;
    getString += " HTTP/1.1\r\n";
    getString += "host: craigandheather.net:80\r\n\r\n";

    // Serial.printf("GET string: %s\n", getString.c_str());

    // Send the GET request to the server
    client.print(getString.c_str());

    delay(50);

    // Wait until data is available
    while (client.available() == 0) {
      delay(50);
    }

    // Start playback
    pVS1053->startPlayback();

    // Raise volume
    pVS1053->setVolume(vol);

    return true;
  }

  void stopPlayback(void) {

    // Stop the VS1053
    pVS1053->stopPlayback();

    delay(50);

    if (client.connected()) {
      // Disconnect from current connection
      client.stop();
    }
    delay(50);
  }

  // Pause playback
  void pausePlayback(boolean state) {
    pVS1053->pausePlayback(state);
  }

  // Move the data to the VS1053
  void moveData(void) {

    int dataAvailable = client.available();
    if (dataAvailable > 0) {
      pVS1053->moveData(buffer, client.read(buffer, min(BUFFER_SIZE, dataAvailable)));
    } else {
      delay(50);
    }
  }

  // Now playing screen
  void displayStationNowPlayingScreen(const char *stationName, const char *countryName) {

    char line[100];

    tft.clearScreen();
    tft.drawRect(0, 0, tft.width(), tft.height(), COLOR_BLUE);

    tft.drawCenteredText(calcLineOffset(1), "- Now Playing -");
    tft.setTextColor(COLOR_WHITE, COLOR_BLUE);

    processString(line, (char *)stationName);

    tft.drawCenteredText(calcLineOffset(4), line);

    processString(line, (char *)countryName);
    tft.drawCenteredText(calcLineOffset(6), line);

    tft.setTextColor(COLOR_WHITE, COLOR_BLACK);

    // Display paused status
    displayPausedStatus(paused);
  }

  // Update the screen above with paused status
  void displayPausedStatus(bool pausedState) {
    if (pausedState) {
      tft.drawCenteredText(calcLineOffset(10), "Paused");
    } else {
      tft.drawCenteredText(calcLineOffset(10), "        ");
    }
  }

  // Perform any setup required for Internet playback
  void irSetup() {

    paused = false;

    // Set the initial state machne start state
    state = GENRE_POPULATE_LB;
  }

  // Realtime loop implementing FSM
  void irLoop() {

    // Run a Finite State Machine (FSM) to control the music player
    switch (state) {

      case GENRE_POPULATE_LB:
        {
          // Display the menu ListBox
          listBox.setDataSource(ARTISTS_DS);
          listBox.clear();
          listBox.setCenterFlag(true);
          listBox.setSortFlag(true);
          listBox.setTitle("- Genre -");

          // Populate list box with genres
          for (int i = 0; i < NUMBER_OF_GENRES; i++) {
            listBox.addEntry(genres[i]);
          }
          // Sort the genres
          listBox.prepareEntries();

          // Display the genres
          listBox.doRepaint(true);

          // Next state
          state = GENRE_NAV_CHECK;
        }
        break;

      case GENRE_NAV_CHECK:
        {
          // Poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          if (result == Navigation::NAV_UP) {
            listBox.selectionDown(true);
          }

          else if (result == Navigation::NAV_DOWN) {
            listBox.selectionUp(true);
          }

          else if (result == Navigation::NAV_BACK) {
            // Nothing to do here
          }

          else if (result == Navigation::NAV_SELECT) {

            // Next state
            state = STATION_POPULATE_LB;
          }
        }
        break;

      case STATION_POPULATE_LB:
        {
          const char *genre = listBox.getSelection();

          // The following is necessary because calling gatherStations
          // somehow kills the genre string. So we copy it before
          // that happens

          // Buffer for holding genre string
          char gBuffer[20];
          strcpy(gBuffer, genre);

          // Serial.printf("Genre: %s\n", gBuffer);

          // Get the stations with this genre
          stationGatherer.gatherStations(genre);

          // Might never get here if radio station provider is offline

          // Get number of returned stations
          int stationCount = stationGatherer.getNumberOfStations();

          Serial.printf("Station count: %d\n", stationCount);

          // Make sure we only add valid stations
          int stationsToList = min(MAX_STATION_COUNT, stationCount);

          // Prepare listbox
          listBox.setDataSource(ALBUMS_DS);
          listBox.clear();
          listBox.setTitle("- Stations -");
          listBox.setCenterFlag(true);

          char snBuffer[30];
          // Assign station names as the genre followed by a number 0 .. 9
          for (int i = 0; i < stationsToList; i++) {
            sprintf(snBuffer, "%s %d", gBuffer, i);
            listBox.addEntry(std::string(snBuffer));
          }

          // Display the stations
          listBox.doRepaint(true);

          // Next state
          state = STATION_NAV_CHECK;
        }
        break;

      case STATION_NAV_CHECK:
        {
          // Poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          if (result == Navigation::NAV_DOWN) {
            listBox.selectionUp(true);
          }

          else if (result == Navigation::NAV_UP) {
            listBox.selectionDown(true);
          }

          else if (result == Navigation::NAV_BACK) {
            // Back to genre selection
            listBox.setDataSource(ARTISTS_DS);
            listBox.doRepaint(true);

            // Next state
            state = GENRE_NAV_CHECK;
          }

          else if (result == Navigation::NAV_SELECT) {
            // A station has been selected

            // Next state
            state = STATION_PLAY;
          }
        }
        break;

      case STATION_PLAY:
        {
          // Get index of selected radio station
          int selectIndex = listBox.getSelectionIndex();

          // We will use the synthesized station names instead of the real names
          // because they are a mess of seemingly bogus characters making them
          // useless. Uncomment the next line and comment out the one after if
          // want to use the real names.
          // String stationName = stationGatherer.getStationNameFromIndex(selectIndex);
          String stationName = listBox.getSelection();
          String stationUrl = stationGatherer.getStationUrlFromIndex(selectIndex);
          String stationCountry = stationGatherer.getStationCountryFromIndex(selectIndex);

          // Show station playing
          displayStationNowPlayingScreen(stationName.c_str(), stationCountry.c_str());

          // Play the station
          startPlayback(stationUrl.c_str());

          // Next state
          state = STATION_STATUS_CHECK;
        }
        break;

      case STATION_STATUS_CHECK:
        {
          // Song is playing so poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          if (result == Navigation::NAV_UP) {
            stopPlayback();

            listBox.selectionDown(false);

            // Next state
            state = STATION_PLAY;
          }

          else if (result == Navigation::NAV_DOWN) {
            stopPlayback();

            listBox.selectionUp(false);

            // Next state
            state = STATION_PLAY;
          }

          else if (result == Navigation::NAV_BACK) {
            stopPlayback();

            // Back to station selection
            listBox.setDataSource(ALBUMS_DS);
            listBox.doRepaint(true);

            // Next state
            state = STATION_NAV_CHECK;
          }

          else if (result == Navigation::NAV_SELECT) {
            // Select button during radio station playback toggles pausing
            paused = !paused;

            // Pause or unpause playback
            pausePlayback(paused);

            displayPausedStatus(paused);
          }
        }
        break;
    }
  }

protected:

  // Pointer to VS1053
  VS1053 *pVS1053;

  boolean paused;

  URL_STRUCT urlStruct;

  // Instantiate WiFi client
  WiFiClient client;

  // Buffer size
  static const int BUFFER_SIZE = VS1053_CHUNK_SIZE * 100;

  byte buffer[BUFFER_SIZE];

  // Instantiate station gatherer
  StationGatherer stationGatherer;

  // Number of stations of specified genre shown in list box
  static const int MAX_STATION_COUNT = 10;

  // Radio station index storage
  int stationIndices[MAX_STATION_COUNT];

  // Finite State Machine states
  enum STATES {

    // Genre states
    GENRE_POPULATE_LB,
    GENRE_NAV_CHECK,

    // Station states
    STATION_POPULATE_LB,
    STATION_NAV_CHECK,
    STATION_PLAY,
    STATION_STATUS_CHECK
  };

  // Current FSM stat
  enum STATES state;

  static const int NUMBER_OF_GENRES = 25;

  // Array of genre names
  const char *genres[NUMBER_OF_GENRES]{
    "60s", "70s", "80s", "90s",
    "metal", "rock", "pop", "disco",
    "trance", "techno", "house", "dance",
    "reggae", "funk", "top", "synth",
    "fantasy", "rain", "ambient", "meditation",
    "lounge", "chill", "classical", "jazz",
    "surf"
  };
};

#endif
