/*
   ListBox's non-GUI Component for ESP32 Music Player

   Attempting to refine repaint mechanism for faster UI updates
   Also attempting to reduce memory footprint substantially

   Concept, design and implementation by: Craig A. Lindley
   Last Update: 12/15/2023
*/

#ifndef LISTBOX_H
#define LISTBOX_H

#include <string>
#include <vector>

#ifndef min
#define min(a, b) (((a) < (b)) ? (a) : (b))
#endif

// Macros for handling random play flags
#define bitRead64(value, bit) (((value) >> (bit)) & 0x01)
#define bitSet64(value, bit) ((value) |= (1ULL << (bit)))

// How many time we will attempt to generate a random select index
// before clearing the flags and trying again
#define MAX_SPINS 45

// Callback is a pointer to a application function that takes two arguments and
// repaints the GIU when called.
typedef void (*callback)(boolean, uint32_t);

// Abstract base class for all data sources
class DataBase {
public:

  void clear(void) {
    selectIndex = 0;
    prevSelectIndex = 0;
    windowIndex = 0;
    centerFlag = false;
    sortFlag = false;

    // Clear title of listbox
    title = "";

    // Clear content of string storage
    entries.clear();
  }

  void setCenterFlag(boolean flag) {
    centerFlag = flag;
  }

  boolean getCenterFlag(void) {
    return centerFlag;
  }

  void setSortFlag(boolean flag) {
    sortFlag = flag;
  }

  boolean getSortFlag(void) {
    return sortFlag;
  }

  void setTitle(std::string theTitle) {
    title = theTitle;
  }

  std::string getTitle(void) {
    return title;
  }

  void addEntry(std::string entry) {
    entries.push_back(entry);
  }

  std::string getEntry(int index) {
    return entries.at(index);
  }

  void prepareEntries(void) {
    if (sortFlag) {
      // Sort the entries in the vector
      sort(entries.begin(), entries.end());
    }
  }

  int getEntryCount(void) {
    return entries.size();
  }

  std::string getSelection(void) {
    return entries.at(selectIndex);
  }

  int getSelectionIndex(void) {
    return selectIndex;
  }

  void selectionUp(boolean repaint) {

    boolean fullRepaint = true;
    int storageIndex = entries.size();

    // Is there an item above ?
    if (selectIndex > 0) {
      // Yes so move selectIndex up
      // Will moving up keep selection in window?
      if (selectIndex > windowIndex) {
        prevSelectIndex = selectIndex;
        selectIndex--;
        fullRepaint = false;
        // No window has to move
      } else if (windowIndex > 0) {
        windowIndex--;
        selectIndex--;
      }
    } else {
      // No, must reposition to end of list
      selectIndex = storageIndex - 1;
      windowIndex = storageIndex - numberOfLines;
      if (windowIndex < 0) {
        windowIndex = 0;
      }
    }
    if (repaint) {
      doRepaint(fullRepaint);
    }
  }

  void selectionDown(boolean repaint) {

    boolean fullRepaint = true;
    int storageIndex = entries.size();

    // If not at end of storage already
    if (selectIndex < storageIndex - 1) {

      // Will moving down keep selection in window?
      if ((selectIndex >= windowIndex) && (selectIndex < (windowIndex + numberOfLines - 1))) {
        prevSelectIndex = selectIndex;
        selectIndex++;
        fullRepaint = false;

        // No window has to move
      } else if ((windowIndex + numberOfLines) < storageIndex) {
        windowIndex++;
        selectIndex++;
      }
    } else {
      selectIndex = 0;
      windowIndex = 0;
    }
    if (repaint) {
      doRepaint(fullRepaint);
    }
  }

  void doRepaint(boolean fullRepaint) {
    repaint(fullRepaint,
            prevSelectIndex << 24 | selectIndex << 16 | windowIndex << 8 | min(entries.size(), numberOfLines));
  }

protected:

  int numberOfLines;
  callback repaint;

  std::string title;
  std::vector<std::string> entries;
  int selectIndex;
  int prevSelectIndex;
  int windowIndex;
  boolean centerFlag;
  boolean sortFlag;
};

class OperationsDS : public DataBase {
public:
  OperationsDS(int _numberOfLines, callback _repaint) {
    numberOfLines = _numberOfLines;
    repaint = _repaint;
  }
};

class ArtistsDS : public DataBase {
public:
  ArtistsDS(int _numberOfLines, callback _repaint) {
    numberOfLines = _numberOfLines;
    repaint = _repaint;
  }
};

class AlbumsDS : public DataBase {
public:
  AlbumsDS(int _numberOfLines, callback _repaint) {
    numberOfLines = _numberOfLines;
    repaint = _repaint;
  }
};

class SongsDS : public DataBase {

public:
  SongsDS(int _numberOfLines, callback _repaint) {
    numberOfLines = _numberOfLines;
    repaint = _repaint;
  }

  void clear() {
    // Do base class clear
    DataBase::clear();

    // Clear flags variable
    flags = 0;
    spins = 0;
  }

  // Select a random entry from the list box
  // The code below tries to play all songs in a directory randomly
  // before repeating any.
  void selectRandomEntry(boolean repaint) {
    int storageIndex = entries.size();

    // Log the song just played
    bitSet64(flags, selectIndex);

    boolean done = false;
    while (!done) {
      // Random count of steps to scroll down in list box
      int steps = random(storageIndex);
      for (int i = 0; i < steps; i++) {
        selectionDown(false);
      }
      // Check to see if the selected song index has been selected before
      if (bitRead64(flags, selectIndex) == 0) {
        // No it hasn't so indicate it has now
        bitSet64(flags, selectIndex);
        // Reset number of spins because new song @ selectIndex has been selected
        spins = 0;
        done = true;
      } else {
        // The selected song index has been selected before, so try again
        // If we have exceeded max attempts then reset flags and start random
        // sequence over.
        if (++spins > MAX_SPINS) {
          flags = 0;
          spins = 0;
        }
      }
    }
    if (repaint) {
      doRepaint(true);
    }
  }

protected:
  // Flags to prevent song repeats when playing songs randomly
  // Maximum of 64 entries supported. See macros at top of file
  uint64_t flags;
  int spins;
};

// Data source identifers for ListBox
enum DATA_SOURCE { OPERATIONS_DS,
                   ARTISTS_DS,
                   ALBUMS_DS,
                   SONGS_DS };

// ListBox class definition
class ListBox {

public:

  // ListBox class constructor
  // numberOfLines is the number of lines in the listbox window
  //   that can be displayed on the display.
  // repaint is a function called whenever the listbox needs updating
  ListBox(int numberOfLines, callback repaint) {

    // Instantiate and initialize each data source
    operationsDS = new OperationsDS(numberOfLines, repaint);
    operationsDS->clear();

    artistsDS = new ArtistsDS(numberOfLines, repaint);
    artistsDS->clear();

    albumsDS = new AlbumsDS(numberOfLines, repaint);
    albumsDS->clear();

    songsDS = new SongsDS(numberOfLines, repaint);
    songsDS->clear();

    // Initialize data source
    setDataSource(OPERATIONS_DS);
  }

  // Select the datasource for the listbox
  void setDataSource(enum DATA_SOURCE ds) {
    switch (ds) {
      case OPERATIONS_DS:
        ptrDS = operationsDS;
        break;
      case ARTISTS_DS:
        ptrDS = artistsDS;
        break;
      case ALBUMS_DS:
        ptrDS = albumsDS;
        break;
      case SONGS_DS:
        ptrDS = songsDS;
        break;
    }
  }
  void selectRandomEntry(boolean repaint) {
    songsDS->selectRandomEntry(repaint);
  }

  void clear(void) {
    ptrDS->clear();
  }

  void setCenterFlag(boolean flag) {
    ptrDS->setCenterFlag(flag);
  }

  boolean getCenterFlag(void) {
    return ptrDS->getCenterFlag();
  }

  void setSortFlag(boolean flag) {
    ptrDS->setSortFlag(flag);
  }

  boolean getSortFlag(void) {
    return ptrDS->getSortFlag();
  }

  void setTitle(std::string title) {
    ptrDS->setTitle(title);
  }

  const char *getTitle(void) {
    return ptrDS->getTitle().c_str();
  }

  void addEntry(std::string entry) {
    ptrDS->addEntry(entry);
  }

  const char *getEntry(int index) {
    return ptrDS->getEntry(index).c_str();
  }

  void prepareEntries(void) {
    ptrDS->prepareEntries();
  }

  int getEntryCount(void) {
    return ptrDS->getEntryCount();
  }

  const char *getSelection(void) {
    return ptrDS->getSelection().c_str();
  }

  int getSelectionIndex(void) {
    return ptrDS->getSelectionIndex();
  }

  void selectionUp(boolean repaint) {
    ptrDS->selectionUp(repaint);
  }

  void selectionDown(boolean repaint) {
    ptrDS->selectionDown(repaint);
  }

  void doRepaint(boolean fullRepaint) {
    ptrDS->doRepaint(fullRepaint);
  }

protected:

  OperationsDS *operationsDS;
  ArtistsDS *artistsDS;
  AlbumsDS *albumsDS;
  SongsDS *songsDS;

  // Pointer to current listbox data source
  DataBase *ptrDS;
};

#endif
