/*
 * This file contains misc code/functions
 *
 * Concept, Design and Implementation by: Craig A. Lindley
 * Last Update: 11/21/2023
 */

uint32_t displayTimeout;

// Calculate y pixel position corresponding to a text line
int calcLineOffset(int line) {
  return line * LINE_HEIGHT;
}

// Determine if a char strings starts with specified prefix
boolean startsWith(const char *pre, const char *str) {

  size_t lenpre = strlen(pre),
         lenstr = strlen(str);
  return lenstr < lenpre ? false : strncmp(pre, str, lenpre) == 0;
}

// This function does two things. First, if the str has a dot followed by
// a file extension, the extension is removed. Second, if the str is too
// long for display, it is truncated to LB_LINE_LENGTH.
// NOTE: Radio station names cannot have periods in them
// or they will be truncated.
void processString(char *line, char *str) {

  memset(line, 0, sizeof(line));
  strcpy(line, str);

  // Search for final . to remove file extension, if any
  char *pDot = strrchr(line, 0x2E);
  if (pDot) {
    // Truncate string at .
    *pDot = '\0';
  }
  // Get the length of the possibly processed string
  int len = strlen(line);

  // Is the string to long to display ?
  if (len > LB_LINE_LENGTH) {
    // Yes, so truncate it
    line[LB_LINE_LENGTH] = '\0';
  }
}

// Function called to repaint list box
void paintListBox(boolean fullRepaint, uint32_t b) {

  char line[80];

  // Extract listbox data from argument
  int prevSelectIndex = (b >> 24) & 0xFF;
  int selectIndex = (b >> 16) & 0xFF;
  int windowIndex = (b >> 8) & 0xFF;
  int numberOfEntries = b & 0xFF;

  // Moves strings over to avoid Rect
  const int xOffset = 5;
  int yOffset = calcLineOffset(1) + 1;

  // Serial.printf("SI: %d, WI: %d, NE: %d\n", selectIndex, windowIndex, numberOfEntries);

  // Fetch the centering flag
  boolean centerFlag = listBox.getCenterFlag();

  if (fullRepaint) {
    // Full repaint

    // First clear the screen
    tft.clearScreen();

    // Draw the enclosing rectangle
    tft.drawRect(0, 0, tft.width(), tft.height(), COLOR_BLUE);

    tft.setTextColor(COLOR_WHITE, COLOR_BLACK);

    // Draw listbox title
    tft.drawCenteredText(calcLineOffset(0), listBox.getTitle());

    for (int i = 0; i < numberOfEntries; i++) {
      // Process the selected entry. See comments above.
      processString(line, (char *)listBox.getEntry(i + windowIndex));

      if ((i + windowIndex) == selectIndex) {
        // This is the selected item. Change its colors
        tft.setTextColor(COLOR_WHITE, COLOR_BLUE);
        if (!centerFlag) {
          tft.drawText(xOffset, yOffset, line);
        } else {
          tft.drawCenteredText(yOffset, line);
        }
      } else {
        // Non selected item
        tft.setTextColor(COLOR_WHITE, COLOR_BLACK);
        if (!centerFlag) {
          // This is a non-selected item
          tft.drawText(xOffset, yOffset, line);
        } else {
          tft.drawCenteredText(yOffset, line);
        }
      }
      // Calculate y offset for next line of display
      yOffset += LINE_HEIGHT;
    }

  } else {

    // Partial repaint
    for (int i = 0; i < numberOfEntries; i++) {
      // Process the selected entry. See comments above.
      processString(line, (char *)listBox.getEntry(i + windowIndex));

      if ((i + windowIndex) == prevSelectIndex) {
        // This is the previously selected item, remove its highlight
        tft.setTextColor(COLOR_WHITE, COLOR_BLACK);
        if (!centerFlag) {
          tft.drawText(xOffset, yOffset, line);
        } else {
          tft.drawCenteredText(yOffset, line);
        }
      }

      if ((i + windowIndex) == selectIndex) {
        // This is the currently selected item. Highlight it
        tft.setTextColor(COLOR_WHITE, COLOR_BLUE);
        if (!centerFlag) {
          tft.drawText(xOffset, yOffset, line);
        } else {
          tft.drawCenteredText(yOffset, line);
        }
      }
      // Calculate y offset for next line of display
      yOffset += LINE_HEIGHT;
    }
  }
  tft.setTextColor(COLOR_WHITE, COLOR_BLACK);
}
