/*
   Rotary Encoder Driver

   Concept, Design and Implementation by: Craig A. Lindley
   Last Update: 11/21/2023
*/

#ifndef ROTARY_ENCODER_H
#define ROTARY_ENCODER_H

class RotaryEncoder {
public:

  // Class constructor
  RotaryEncoder(uint8_t plusPin, uint8_t minusPin) {
    // Save GPIO pin numbers
    _rePlusPin = plusPin;
    _reMinusPin = minusPin;

    // Set pin modes
    pinMode(_rePlusPin, INPUT_PULLUP);
    pinMode(_reMinusPin, INPUT_PULLUP);

    // Set initial state
    _reState = STATE_START;

    // Initial encoder value
    encoderValue = 0;
  }

  // Set encoder value
  void setEncoderValue(uint16_t val) {
    encoderValue = val;
  }

  // Get encoder value
  uint16_t getEncoderValue() {
    return encoderValue;
  }

  // Read and process encoder data
  // Keeping the variable encoderValue within specified bounds
  boolean readEncoder(uint16_t inc, uint16_t minVal, uint16_t maxVal) {

    // Update FSM
    uint8_t fsmVal = updateRotaryEncoder();

    if (fsmVal == FCODE) {
      encoderValue += inc;
      if (encoderValue > maxVal) {
        encoderValue = maxVal;
      }
      return true;

    } else if (fsmVal == BCODE) {
      encoderValue -= inc;
      if (encoderValue < (int16_t)minVal) {
        encoderValue = minVal;
      }
      return true;

    } else {
      // No encoder rotation
      return false;
    }
  }

  // Read and process encoder's rotary direction
  // Return  1 if encoder is moved in positive direction
  // Return -1 if encoder is moved in negative direction
  // Return 0 if no movement
  int readEncoderDirection() {
    // Update FSM
    uint8_t fsmVal = updateRotaryEncoder();

    if (fsmVal == FCODE) {
      return 1;
    } else if (fsmVal == BCODE) {
      return -1;
    } else {
      return 0;
    }
  }

  // This should be called often to update the FSM
  uint16_t updateRotaryEncoder() {

    uint16_t _reValue = 0;

    // Read the encoder
    uint8_t in = readEncoderPins();
    switch (_reState) {

      case STATE_START:
        if (in == 1) {
          _reState = STATE_B1;
        } else if (in == 2) {
          _reState = STATE_F1;
        }
        break;

      case STATE_F1:
        if (in == 3) {
          _reState = STATE_F2;
        } else if (in == 2) {
          // Nothing to do
        } else {
          _reState = STATE_ERROR;
        }
        break;

      case STATE_F2:
        if (in == 1) {
          _reState = STATE_F3;
        } else if (in == 3) {
          // Nothing to do
        } else {
          _reState = STATE_ERROR;
        }
        break;

      case STATE_F3:
        if (in == 0) {
          _reValue = FCODE;
          _reState = STATE_START;
        } else if (in == 1) {
          // Nothing to do
        } else {
          _reState = STATE_ERROR;
        }
        break;

      case STATE_B1:
        if (in == 3) {
          _reState = STATE_B2;
        } else if (in == 1) {
          // Nothing to do
        } else {
          _reState = STATE_ERROR;
        }
        break;

      case STATE_B2:
        if (in == 2) {
          _reState = STATE_B3;
        } else if (in == 3) {
          // Nothing to do
        } else {
          _reState = STATE_ERROR;
        }
        break;

      case STATE_B3:
        if (in == 0) {
          _reValue = BCODE;
          _reState = STATE_START;
        } else if (in == 2) {
          // Nothing to do
        } else {
          _reState = STATE_ERROR;
        }
        break;

      case STATE_ERROR:
        if (in == 0) {
          _reState = STATE_START;
        }
        break;
    }
    return _reValue;
  }

protected:

  enum FSM_STATES {
    STATE_START,
    STATE_F1,
    STATE_F2,
    STATE_F3,
    STATE_B1,
    STATE_B2,
    STATE_B3,
    STATE_ERROR
  };

  // Protected data

  // Current finite state machine state
  enum FSM_STATES _reState;

  uint8_t _rePlusPin, _reMinusPin;
  int16_t encoderValue;

  static const uint8_t FCODE = 0x10;
  static const uint8_t BCODE = 0x20;

  // Protected methods

  // Capture current state of encoder pins
  uint8_t readEncoderPins() {
    uint8_t t = (digitalRead(_rePlusPin) << 1) | digitalRead(_reMinusPin);

    return ~t & 0x03;
  }
};

#endif
