/*
 * This class encapsulates the SD music player Functionality
 *
 * Concept, Design and Implementation by: Craig A. Lindley
 * Last Update: 12/15/2023
 */

#ifndef SD_PLAYER
#define SD_PLAYER

#include "FTPServer.h"

// Credentials for FTP connection for remote access
#define FTP_USER "craig"
#define FTP_PSWD "music"

class SDPlayer {

public:
  // Class Constructor
  SDPlayer(VS1053 *_pVS1053) {
    // Save incoming
    pVS1053 = _pVS1053;
  }

  void startPlayback(const char *filePath) {

    // Open file for playing
    file = SD.open(filePath);
    if (!file) {
      Serial.println("Failed to open file for reading");
      return;
    }

    // Start the VS1053
    pVS1053->startPlayback();
  }

  // Return music playing status
  boolean isPlaying() {
    return pVS1053->isPlaying();
  }

  void stopPlayback(void) {

    // Stop the VS1053
    pVS1053->stopPlayback();

    delay(50);

    // Close the music file
    file.close();
  }

  // Pause playback
  void pausePlayback(boolean state) {
    pVS1053->pausePlayback(state);
  }

  // Move the file data to the VS1053
  void moveData(void) {

    int dataAvailable = file.available();
    if (dataAvailable <= 0) {
      file.close();

      // Stop the VS1053
      pVS1053->stopPlayback();
      return;
    }
    int dataToRead = min(BUFFER_SIZE, dataAvailable);
    int dataRead = file.read(buffer, dataToRead);
    pVS1053->moveData(buffer, dataRead);
  }

  // Uploading screen
  void displayUploadScreen() {

    // Draw outline
    tft.clearScreen();
    tft.drawRect(0, 0, tft.width(), tft.height(), COLOR_BLUE);

    // Draw title
    tft.drawCenteredText(calcLineOffset(2), "Preparing For");
    tft.drawCenteredText(calcLineOffset(3), "Remote Access");

    // Show IP address for FTP
    tft.drawCenteredText(calcLineOffset(4), "IP Address");
    tft.drawCenteredText(calcLineOffset(5), WiFi.localIP().toString().c_str());
  }

  // Now playing screen
  void displaySongNowPlayingScreen(const char *artistName, const char *albumName, const char *songName) {

    char line[40];

    processString(line, (char *)songName);

    tft.clearScreen();
    tft.drawRect(0, 0, tft.width(), tft.height(), COLOR_BLUE);

    tft.drawCenteredText(calcLineOffset(1), "- Now Playing -");
    tft.drawCenteredText(calcLineOffset(3), artistName);
    tft.drawCenteredText(calcLineOffset(4), albumName);
    tft.setTextColor(COLOR_WHITE, COLOR_BLUE);
    tft.drawCenteredText(calcLineOffset(6), line);
    tft.setTextColor(COLOR_WHITE, COLOR_BLACK);

    // Display looping status
    displayLoopStatus(looping);
  }

  // Update the screen above with loop status
  void displayLoopStatus(bool loopSet) {
    if (loopSet) {
      tft.drawCenteredText(calcLineOffset(10), "Looping");
    } else {
      tft.drawCenteredText(calcLineOffset(10), "        ");
    }
  }

  // Perform any setup required for SD playback
  void sdSetup() {

    // Initialize control variables
    looping = false;
    skipInput = false;
    uploading = false;

    // Set the initial FSM start state
    state = OP_POPULATE_LB;
  }

  // Realtime loop implementing FSM
  void sdLoop() {

    if (uploading) {
      // Feed the FTP server
      ftpServer.handleFTP();
    }

    // Run a Finite State Machine (FSM) to control the music player
    switch (state) {

      case OP_POPULATE_LB:
        {
          // Display the menu ListBox
          listBox.setDataSource(OPERATIONS_DS);
          listBox.clear();
          listBox.setCenterFlag(true);
          listBox.setTitle("- Operations -");
          listBox.addEntry("Sequential Play");
          listBox.addEntry("Random Play");
          listBox.addEntry("Remote Access");

          listBox.doRepaint(true);

          // Next state
          state = OP_NAV_CHECK;
        }
        break;

      case OP_NAV_CHECK:
        {
          // Poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          if (result != Navigation::NAV_NONE) {
            updateTimeOut();
          }

          if (result == Navigation::NAV_UP) {
            listBox.selectionDown(true);
          }

          else if (result == Navigation::NAV_DOWN) {
            listBox.selectionUp(true);
          }

          else if (result == Navigation::NAV_BACK) {
            // Nothing to do here
          }

          else if (result == Navigation::NAV_SELECT) {

            // Next state
            state = OP_DISPATCH;
          }
        }
        break;

      case OP_DISPATCH:
        {
          switch (listBox.getSelectionIndex()) {
            case 0:
              // Sequential play selected
              playMode = SEQUENTIAL;

              // Next state
              state = AR_POPULATE_LB;
              break;

            case 1:
              // Random play selected
              playMode = RANDOM;

              // Next state
              state = AR_POPULATE_LB;
              break;

            case 2:
              // Remote access selected

              // Next state
              state = RA_WIFI_CONNECT;
              break;
          }
        }
        break;

      case AR_POPULATE_LB:
        {
          // Root directory path
          strcpy(songPath, "/");

          listBox.setDataSource(ARTISTS_DS);
          listBox.clear();
          listBox.setTitle("- Artists -");
          listBox.setCenterFlag(true);
          listBox.setSortFlag(true);

          // Open the root directory of SD card
          File root = SD.open(songPath);
          if (!root) {
            Serial.println("Failed to open directory");
            return;
          }

          // Populate listbox with directory names
          File file = root.openNextFile();
          while (file) {
            if (file.isDirectory()) {
              const char *name = file.name();
              // Ignore the pesky hidden files
              if (isAlphaNumeric(name[0])) {
                listBox.addEntry(name);
              }
            }
            file.close();
            file = root.openNextFile();
          }
          // Close the files
          file.close();
          root.close();

          // Sort the artists
          listBox.prepareEntries();

          // Display the artists
          listBox.doRepaint(true);

          // Next state
          state = AR_NAV_CHECK;
        }
        break;

      case AR_NAV_CHECK:
        {
          // Poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          if (result != Navigation::NAV_NONE) {
            updateTimeOut();
          }

          if (result == Navigation::NAV_DOWN) {
            listBox.selectionUp(true);
          }

          else if (result == Navigation::NAV_UP) {
            listBox.selectionDown(true);
          }

          else if (result == Navigation::NAV_BACK) {
            // Null out song path
            *songPath = '\0';

            // Back to operation selection
            listBox.setDataSource(OPERATIONS_DS);
            listBox.doRepaint(true);

            // Next state
            state = OP_NAV_CHECK;
          }

          else if (result == Navigation::NAV_SELECT) {
            // An artist has been selected
            // Save the artist's name
            // Clip to list box dimensions
            char line[40];
            processString(line, (char *)listBox.getSelection());
            strcpy(currentArtist, line);

            // Next state
            state = AL_POPULATE_LB;
          }
        }
        break;

      case AL_POPULATE_LB:
        {
          // Add artist to song path
          strcat(songPath, listBox.getSelection());

          listBox.setDataSource(ALBUMS_DS);
          listBox.clear();
          listBox.setTitle("- Albums/CDs -");
          listBox.setCenterFlag(true);
          listBox.setSortFlag(true);

          // Open the artist's directory
          File root = SD.open(songPath);
          if (!root) {
            Serial.println("Failed to open directory");
            return;
          }

          // Populate listbox with album names
          File file = root.openNextFile();
          while (file) {
            if (file.isDirectory()) {
              const char *name = file.name();
              // Ignore the pesky hidden files
              if (isAlphaNumeric(name[0])) {
                listBox.addEntry(name);
              }
            }
            file.close();
            file = root.openNextFile();
          }
          // Close the files
          file.close();
          root.close();

          // Sort the albums
          listBox.prepareEntries();

          // Display the albums
          listBox.doRepaint(true);

          // Next state
          state = AL_NAV_CHECK;
        }
        break;

      case AL_NAV_CHECK:
        {
          // Poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          if (result != Navigation::NAV_NONE) {
            updateTimeOut();
          }

          if (result == Navigation::NAV_UP) {
            listBox.selectionDown(true);
          }

          else if (result == Navigation::NAV_DOWN) {
            listBox.selectionUp(true);
          }

          else if (result == Navigation::NAV_BACK) {
            // Root directory path
            strcpy(songPath, "/");

            // Back to artist selection
            listBox.setDataSource(ARTISTS_DS);
            listBox.doRepaint(true);

            // Next state
            state = AR_NAV_CHECK;
          }

          else if (result == Navigation::NAV_SELECT) {
            // And album has been selected
            // Save the album's name
            // Clip to list box dimensions
            char line[40];
            processString(line, (char *)listBox.getSelection());
            strcpy(currentAlbum, line);

            // Next state
            state = SG_POPULATE_LB;
          }
        }
        break;

      case SG_POPULATE_LB:
        {
          // Add album to song path
          strcat(songPath, "/");
          strcat(songPath, listBox.getSelection());

          listBox.setDataSource(SONGS_DS);
          listBox.clear();
          listBox.setTitle("- Songs -");
          listBox.setCenterFlag(true);
          listBox.setSortFlag(true);

          // Open the album's directory
          File root = SD.open(songPath);
          if (!root) {
            Serial.println("Failed to open directory");
            return;
          }

          // Populate listbox with song filenames
          File file = root.openNextFile();
          while (file) {
            if (!file.isDirectory()) {
              const char *name = file.name();
              // Ignore the pesky hidden files
              if (isAlphaNumeric(name[0])) {
                listBox.addEntry(name);
              }
            }
            file.close();
            file = root.openNextFile();
          }
          // Close the files
          file.close();
          root.close();

          // Sort the songs
          listBox.prepareEntries();

          // Display the songs
          listBox.doRepaint(true);

          // Next state
          state = SG_NAV_CHECK;
        }
        break;

      case SG_NAV_CHECK:
        {
          // Poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          if (result != Navigation::NAV_NONE) {
            updateTimeOut();
          }

          if (result == Navigation::NAV_UP) {
            listBox.selectionDown(true);
          }

          else if (result == Navigation::NAV_DOWN) {
            listBox.selectionUp(true);
          }

          else if (result == Navigation::NAV_BACK) {
            // Find last forward slash in file path
            char *lastSlash = strrchr(songPath, 0x2F);

            // Terminate song path there
            *lastSlash = '\0';

            // Back to album selection
            listBox.setDataSource(ALBUMS_DS);
            listBox.doRepaint(true);

            // Next state
            state = AL_NAV_CHECK;
          }

          else if (result == Navigation::NAV_SELECT) {
            // An song has been selected

            // Next state
            state = SG_PLAY;
          }
        }
        break;

      case SG_PLAY:
        {
          // Add song filename to song path
          strcat(songPath, "/");
          strcat(songPath, listBox.getSelection());

          // Serial.printf("File to play: %s\n", songPath);

          // Display the song to play
          displaySongNowPlayingScreen(currentArtist, currentAlbum, listBox.getSelection());

          // Turn display back on if off for song change
          updateTimeOut();

          startPlayback(songPath);

          // Next state
          state = SG_SONGSTATUS_CHECK;
        }
        break;

      case SG_SONGSTATUS_CHECK:
        {
          // Has the song ended ?
          if (!isPlaying()) {
            // Song has ended
            // Are we looping on this song ?
            if (looping) {
              // Looping
            } else {
              // Not looping
              if (playMode == SEQUENTIAL) {
                listBox.selectionDown(false);
              } else {
                listBox.selectRandomEntry(false);
              }
            }
            // Next state
            state = SG_PATH_RESET;
            break;
          }
          // Song is playing so poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          // If skipInput is true we are trying to consume
          // the first button press because it should just
          // turn the backlight back on

          if (skipInput) {
            if (result != Navigation::NAV_NONE) {
              skipInput = false;
              updateTimeOut();

              // Stay in this state
              break;
            }
          }
          if (result != Navigation::NAV_NONE) {
            updateTimeOut();
          }

          if (result == Navigation::NAV_UP) {
            stopPlayback();

            listBox.selectionDown(false);

            // Next state
            state = SG_PATH_RESET;
          }

          else if (result == Navigation::NAV_DOWN) {
            stopPlayback();

            listBox.selectionUp(false);

            // Next state
            state = SG_PATH_RESET;
          }

          else if (result == Navigation::NAV_BACK) {
            stopPlayback();

            // Back to song selection
            listBox.setDataSource(SONGS_DS);
            listBox.doRepaint(true);

            // Remove previous song from song path
            char *lastSlash = strrchr(songPath, 0x2F);
            *lastSlash = '\0';

            // Next state
            state = SG_NAV_CHECK;
          }

          else if (result == Navigation::NAV_SELECT) {
            // Select button during song playback toggles looping
            looping = !looping;

            displayLoopStatus(looping);
          }

          else if (millis() > displayTimeout) {
            // If skipInput is true
            // Ignore the button press that turns the display backlight on
            skipInput = true;
            tft.setBacklight(LOW);
          }
        }
        break;

      case SG_PATH_RESET:
        {
          // Remove previous song from song path
          char *lastSlash = strrchr(songPath, 0x2F);
          *lastSlash = '\0';

          // Next state
          state = SG_PLAY;
        }
        break;

      case RA_WIFI_CONNECT:
        {
          // Initialize FTP server
          ftpServer.begin(FTP_USER, FTP_PSWD);

          // Next state
          state = RA_DISPLAY;
        }
        break;

      case RA_DISPLAY:
        {
          displayUploadScreen();

          // Indicating we are uploading
          uploading = true;

          // Next state
          state = RA_NAV_CHECK;
        }
        break;

      case RA_NAV_CHECK:
        {
          // Poll the songs navigational rotary encoder
          // Any action cancels FTP
          byte result = songsNav.getNavigationDirection();

          if (result != Navigation::NAV_NONE) {
            // Indicate not uploading
            uploading = false;

            // Do complete restart
            ESP.restart();
          }
        }
        break;
    }
  }

protected:

  // Buffer size
  static const int BUFFER_SIZE = VS1053_CHUNK_SIZE * 100;
  byte buffer[BUFFER_SIZE];

  VS1053 *pVS1053;

  File file;

  boolean looping;
  boolean skipInput;
  boolean uploading;

  // Declare FTP server instance
  FTPServer ftpServer;

  // Buffer for building paths to song files on SD card
  char songPath[100];

  enum PLAY_MODE { SEQUENTIAL,
                   RANDOM };
  enum PLAY_MODE playMode;

  // Storage for artist and album
  char currentArtist[30];
  char currentAlbum[30];

  // Finite State Machine states
  enum STATES {

    // Operation states
    OP_POPULATE_LB,
    OP_NAV_CHECK,
    OP_DISPATCH,

    // Artist states
    AR_POPULATE_LB,
    AR_NAV_CHECK,

    // Album states
    AL_POPULATE_LB,
    AL_NAV_CHECK,

    // Song states
    SG_POPULATE_LB,
    SG_NAV_CHECK,
    SG_PLAY,
    SG_SONGSTATUS_CHECK,
    SG_PATH_RESET,

    // Remote access states
    RA_WIFI_CONNECT,
    RA_DISPLAY,
    RA_NAV_CHECK,
  };

  // FSM state
  STATES state;
};

#endif
