/*
 * This code gathers data on Internet Radio Stations
 * based on their genre, codec and bitrate.
 *
 * Concept, Design and Implementation by: Craig A. Lindley
 * Last Update: 12/27/2023
 */

#ifndef STATION_GATHERER
#define STATION_GATHERER

#include "JsonStreamingParser.h"  // https://github.com/squix78/json-streaming-parser
#include "AJsonListener.h"

#define IR_SERVER1 "at1.api.radio-browser.info"
#define IR_SERVER2 "de1.api.radio-browser.info"
#define IR_SERVER3 "nl1.api.radio-browser.info"

class StationGatherer {

public:

  // Class constructor
  StationGatherer() {
    // Install Json listener
    parser.setListener(&listener);
  }

  // Gather Internet Radio Stations of specified genre
  void gatherStations(const char *genre) {

    // Seems to be necessary to prevent reboot when genre changed
    delay(10);

    // Reset the parser
    parser.reset();

    // Clear any previous data in storage vectors
    listener.clearVectorData();

    // If client was connected, disconnect
    if (client.connected()) {
      // Disconnect from current connection
      client.stop();
    }

    // Attempt connection to Internet Radio data repository
    if (!client.connect(IR_SERVER1, 80)) {
      Serial.println("\nNetwork connection failed");
      return;
    }

    // This should not be necessary but it seems to be
    String genreString = String(genre);

    // Form the GET string. Host must be specified or
    // server won't recognize the request
    String getString = String("GET");
    getString += " /json/stations/search?hidebroken=true&tag=";
    getString += genreString;
    getString += "&codec=MP3&bitrateMax=192&offset=0&limit=10";
    getString += " HTTP/1.1\r\n";
    getString += "host: craigandheather.net:80\r\n\r\n";

    // Serial.printf("GET string: %s\n", getString.c_str());

    // Send the GET request to the server
    client.print(getString.c_str());

    // Wait until data is available
    while (client.available() == 0) {
      delay(50);
    }

    char ch;

    // Consume HTTP header
    while ((ch = client.read()) != '[') {
      Serial.print(ch);
    }

    // Serial.println("H3");

    // Inject first character of Json data into parser
    parser.parse('[');

    while (client.available() > 0) {
      ch = client.read();
      parser.parse(ch);
    }
    client.stop();
  }

  int getNumberOfStations() {
    return listener.getNumberOfStations();
  }

  // Get the Internet Radio Station's name from its index
  String getStationNameFromIndex(int index) {
    return listener.getStationNameFromIndex(index);
  }

  // Get the Internet Radio Station's URL from its index
  String getStationUrlFromIndex(int index) {
    return listener.getStationUrlFromIndex(index);
  }

  // Get the Internet Radio Station's country from its index
  String getStationCountryFromIndex(int index) {
    return listener.getStationCountryFromIndex(index);
  }

protected:
  // Instantiate WiFi client
  WiFiClient client;

  // Instantiate Json listener
  AJsonListener listener;

  // Instantiate Json parser
  JsonStreamingParser parser;
};

#endif
