/*
  ESP32 Driver for a generic VS1053 module

  Non Interrupt Driven

  Concept, Design and Implementation: Craig A. Lindley
  Last Update: 11/21/2023
*/

#ifndef VS1053_H
#define VS1053_H

#undef ENABLE_TONE_CONTROLS 

#define DEFAULT_VOLUME 90

#define VS1053_CHUNK_SIZE 32

class VS1053 {

public:

  // Initialize the VS1053 for operation
  void begin(void) {

    // Do hardware initialization
    hardwareInit();

    // Initialize configuration data basically to off
    ConfigData.playMode = true;  // False is sequential play mode
    ConfigData.volume = DEFAULT_VOLUME;
    ConfigData.trebleLevel = 0;
    ConfigData.trebleFreq = 10;
    ConfigData.bassLevel = 0;
    ConfigData.bassFreq = 15;
    ConfigData.earSpeakerMode = 0;

    playing = false;

    // Initialize SPI hardware for slow operation
    spiSettings = SPISettings(1000000, MSBFIRST, SPI_MODE0);

    // Verify a VS1053 is connected
    if (!verifyVS1053()) {
      Serial.println("Bad or missing VS1053");
      while (1) {}
    }

    // Perform soft reset
    softReset();

    // Set 44.1 KHz stereo
    writeReg(SCI_AUDATA, 44101);

    // Set clock speedup to 3x
    writeReg(SCI_CLOCKF, 0x6000);

    // Now initialize SPI hardware for fast operation
    spiSettings = SPISettings(8000000, MSBFIRST, SPI_MODE0);

    // Wait for a data request before continuing
    awaitDataRequest();

    // Set default volume
    setVolume(DEFAULT_VOLUME);

    Serial.println("VS1053 Ready");
  }

  // Start playback of audio stream
  void startPlayback(void) {
    // Await for data request
    awaitDataRequest();

    playing = true;
  }

  // Return music playing status
  boolean isPlaying() {
    return playing;
  }

  // Stop playback
  void stopPlayback(void) {
    playing = false;
    softReset();
  }

  // Pause playback
  void pausePlayback(boolean state) {
    if (state) {
      playing = false;
    } else {
      playing = true;
    }
  }

  // Set playback volume
  // Range: 0 .. 100 with 100 being the loudest
  void setVolume(int vol) {
    if (vol < 0) {
      vol = 0;
    } else if (vol > 100) {
      vol = 100;
    }

    // Save new volume
    ConfigData.volume = vol;

    // Use map to change positive values to -0.5 dB steps
    uint16_t val = map(vol, 0, 100, 0xFE, 0x00);

    // Must set volume separately for left and right channels
    uint16_t value = (val << 8) | val;
    writeReg(SCI_VOL, value);
    delay(20);
  }

  int getVolume() {
    return ConfigData.volume;
  }

  // Sinewav test tone
  // See vs1053.pdf pg 67 for info
  // n == 3 should result in a sinewave of
  // frequency 1033 Hz
  void sineTest(uint8_t n, uint16_t ms) {

    softReset();

    uint16_t mode = readReg(SCI_MODE);
    mode |= 0x0020;
    writeReg(SCI_MODE, mode);

    awaitDataRequest();

    uint8_t sine_start[8] = { 0x53, 0xEF, 0x6E, n, 0x00, 0x00, 0x00, 0x00 };
    uint8_t sine_stop[8] = { 0x45, 0x78, 0x69, 0x74, 0x00, 0x00, 0x00, 0x00 };

    dataModeOn();

    SPI.writeBytes(sine_start, 8);
    delay(ms);
    SPI.writeBytes(sine_stop, 8);

    dataModeOff();
  }

#ifdef ENABLE_TONE_CONTROLS

  // Return the current treble amplitude
  // in range -8 to +7 dB with 0 being off
  int getTrebleLevel(void) {
    return ConfigData.trebleLevel;
  }

  // Set the treble amplitude
  // -8 <= amplitude <= 7 dB in 1.5 dB steps with 0 being off
  void setTrebleLevel(int amplitude) {
    if (amplitude < -8) {
      amplitude = -8;
    } else if (amplitude > 7) {
      amplitude = 7;
    }
    ConfigData.trebleLevel = amplitude;
    configureEQ();
  }

  // Return the current treble frequency limit
  // in range 1000 Hz to 15,000 Hz
  int getTrebleFreq(void) {
    return ConfigData.trebleFreq * 1000;
  }

  // Set the treble frequency limit
  // in range 1000 Hz to 15,000 Hz in 1000 Hz steps
  void setTrebleFreq(int frequency) {
    frequency /= 1000;

    if (frequency < 1) {
      frequency = 1;
    } else if (frequency > 15) {
      frequency = 15;
    }
    ConfigData.trebleFreq = frequency;
    configureEQ();
  }

  // Return the current bass boost amplitude
  // in range 0 to 15 dB
  int getBassLevel(void) {
    return ConfigData.bassLevel;
  }

  // Set the bass boost amplitude
  // in range 0 to 15 dB in 1 dB steps
  void setBassLevel(int amplitude) {
    if (amplitude < 0) {
      amplitude = 0;
    } else if (amplitude > 15) {
      amplitude = 15;
    }
    ConfigData.bassLevel = amplitude;
    configureEQ();
  }

  // Return the current bass frequency limit
  // in range 20 Hz to 150 Hz
  int getBassFreq(void) {
    return ConfigData.bassFreq * 10;
  }

  // Set the bass frequency limit
  // in range 20 Hz to 150 Hz in 10 Hz steps
  void setBassFreq(int frequency) {
    frequency /= 10;
    if (frequency < 2) {
      frequency = 2;
    } else if (frequency > 15) {
      frequency = 15;
    }
    ConfigData.bassFreq = frequency;
    configureEQ();
  }

  // Get the current earSpeaker mode
  // in range 0 to 3
  int getEarSpeakerMode(void) {
    return ConfigData.earSpeakerMode;
  }

  // Set the ear speaker mode for headphone listening
  // mode is 0 for off; 1 for minimal; 2 for normal; 3 for extreme
  void setEarSpeakerMode(int esMode) {

    // Force mode into valid range
    esMode %= 4;

    // Save setting
    ConfigData.earSpeakerMode = esMode;

    // Bits controlling ear speaker mode
    uint16_t SM_LO = 1 << 4;
    uint16_t SM_HI = 1 << 7;

    // Read the current mode register value
    uint16_t modeReg = readReg(SCI_MODE);

    // Turn off both mode bits
    modeReg &= 0xFF6F;

    switch (esMode) {
      // Mode off
      case 0:
        // Nothing to do
        break;

      // Mode minimal
      case 1:
        modeReg |= SM_LO;
        break;

      // Mode normal
      case 2:
        modeReg |= SM_HI;
        break;

      // Mode extreme
      case 3:
        modeReg |= SM_LO;
        modeReg |= SM_HI;
        break;
    }
    writeReg(SCI_MODE, modeReg);
  }

#endif // ENABLE_TONE_CONTROLS

  // Move data to the VS1053 via SPI
  void moveData(byte *buffer, int count) {

    // Length of chunk 32 byte or shorter
    int chunkLength;

    dataModeOn();
    while (count) {
      chunkLength = count;
      if (count > VS1053_CHUNK_SIZE) {
        chunkLength = VS1053_CHUNK_SIZE;
      }
      count -= chunkLength;

      // Wait for space available
      awaitDataRequest();
      SPI.writeBytes(buffer, chunkLength);
      buffer += chunkLength;
    }
    dataModeOff();
  }

protected:

  // Do a soft reset of the VS1053
  void softReset(void) {
    writeReg(SCI_MODE, (1 << SM_SDINEW) | (1 << SM_RESET));
    delay(100);
  }

  inline void awaitDataRequest() {
    while (!digitalRead(VS_DREQ)) {}
  }

  inline void controlModeOn() {
    SPI.beginTransaction(spiSettings);  // Prevent other SPI users
    digitalWrite(VS_xDCS, HIGH);        // Bring slave in control mode
    digitalWrite(VS_xCS, LOW);
  }

  inline void controlModeOff() {
    digitalWrite(VS_xCS, HIGH);  // End control mode
    SPI.endTransaction();        // Allow other SPI users
  }

  inline void dataModeOn() {
    SPI.beginTransaction(spiSettings);  // Prevent other SPI users
    digitalWrite(VS_xCS, HIGH);         // Bring slave in data mode
    digitalWrite(VS_xDCS, LOW);
  }

  inline void dataModeOff() {
    digitalWrite(VS_xDCS, HIGH);  // End data mode
    SPI.endTransaction();         // Allow other SPI users
  }

  // Read specified VS1053 register
  uint16_t readReg(int reg) {

    controlModeOn();
    SPI.write(3);  // Read operation
    SPI.write(reg);
    uint16_t result = SPI.transfer16(0xFFFF);

    // Wait for a data request before continuing
    awaitDataRequest();
    controlModeOff();
    return result;
  }

  // Write value to specified VS1053 register
  void writeReg(int reg, uint16_t value) {

    controlModeOn();
    SPI.write(2);  // Write operation
    SPI.write(reg);
    SPI.write16(value);

    // Wait for a data request before continuing
    awaitDataRequest();
    controlModeOff();
  }

  // Hardware initialization
  void hardwareInit() {

    // Configure GPIO lines
    pinMode(VS_DREQ, INPUT);

    // VS1053 control chip select
    pinMode(VS_xCS, OUTPUT);
    digitalWrite(VS_xCS, HIGH);

    // VS1053 data chip select
    pinMode(VS_xDCS, OUTPUT);
    digitalWrite(VS_xDCS, HIGH);

    // Configure and activate reset
    pinMode(VS_RST, OUTPUT);
    digitalWrite(VS_RST, HIGH);
    delay(100);

    // Pull reset line low or active
    digitalWrite(VS_RST, LOW);
    delay(100);

    // Release the reset line
    digitalWrite(VS_RST, HIGH);
    delay(100);
  }

  // Verify that SPI can read and write the volume register
  // without error. Return false on data miscompare.
  bool verifyVS1053(void) {

    // Walk a one across volume level integer
    for (int i = 0; i < 16; i++) {
      int out = 1 << i;
      writeReg(SCI_VOL, out);
      int in = readReg(SCI_VOL);
      if (in != out) {
        return false;
      }
    }
    return true;
  }

  // Configure the EQ
  void configureEQ(void) {
    unsigned short newEQ = ((ConfigData.trebleLevel & 0x0f) << 12)
                           | ((ConfigData.trebleFreq & 0x0f) << 8)
                           | ((ConfigData.bassLevel & 0x0f) << 4)
                           | ((ConfigData.bassFreq & 0x0f) << 0);

    writeReg(SCI_BASS, newEQ);
  }

  // SCI Registers
  const byte SCI_MODE = 0;
  const byte SCI_STATUS = 1;
  const byte SCI_BASS = 2;
  const byte SCI_CLOCKF = 3;
  const byte SCI_DECODE_TIME = 4;
  const byte SCI_AUDATA = 5;
  const byte SCI_WRAM = 6;
  const byte SCI_WRAMADDR = 7;
  const byte SCI_HDAT0 = 8;
  const byte SCI_HDAT1 = 9;
  const byte SCI_AIADDR = 10;
  const byte SCI_VOL = 11;
  const byte SCI_AICTRL0 = 12;
  const byte SCI_AICTRL1 = 13;
  const byte SCI_AICTRL2 = 14;
  const byte SCI_AICTRL3 = 15;

  // SCI_MODE Bits
  const byte SM_DIFF = 0;
  const byte SM_LAYER12 = 1;
  const byte SM_RESET = 2;
  const byte SM_OUTOFWAV = 3;
  const byte SM_EARSPEAKER_LO = 4;
  const byte SM_TESTS = 5;
  const byte SM_STREAM = 6;
  const byte SM_EARSPEAKER_HI = 7;
  const byte SM_DACT = 8;
  const byte SM_SDIORD = 9;
  const byte SM_SDISHARE = 10;
  const byte SM_SDINEW = 11;
  const byte SM_ADPCM = 12;
  const byte SM_ADCPM_HP = 13;
  const byte SM_LINE_IN = 14;
  const byte SM_CLK_RANGE = 15;

  SPISettings spiSettings;
  boolean playing;

  // Configuration data record structure
  struct _configData {
    boolean playMode;
    int volume;
    int trebleLevel;
    int trebleFreq;
    int bassLevel;
    int bassFreq;
    int earSpeakerMode;
  };

  struct _configData ConfigData;
};

#endif
