/*
 * This class encapsulates the network/web music player functionality
 *
 * Concept, Design and Implementation by: Craig A. Lindley
 * Last Update: 12/14/2023
 */

#ifndef WEB_PLAYER
#define WEB_PLAYER

#include "WebScraper.h"

class WEBPlayer {

public:

  // Class constructor
  WEBPlayer(VS1053 *_pVS1053) {
    pVS1053 = _pVS1053;
  }

  // Determine if a song is still playing
  bool isPlaying() {
    return pVS1053->isPlaying();
  }

  bool startPlayback(const char *url) {

    // Get the current volume level
    int vol = pVS1053->getVolume();

    // Lower volume during change of stream
    pVS1053->setVolume(0);

    // Stop playback
    pVS1053->stopPlayback();

    if (client.connected()) {
      // Disconnect from current connection
      client.stop();
      delay(50);
    }

    // Clear audio data buffer
    memset(buffer, 0, sizeof(buffer));

    // Attempt connection to host
    if (!client.connect("craigandheather.net", 80)) {
      Serial.println("\nNetwork connection failed");
      return false;
    }

    delay(50);

    // Form the GET string. Host must be specified or
    // server won't recognize the request
    String getString = String("GET ") + url + " HTTP/1.1\r\n";
    getString += "host: craigandheather.net:80\r\n\r\n";

    // Send the GET request to the server
    client.print(getString.c_str());

    delay(50);

    // Wait until data is available
    while (client.available() == 0) {
      delay(50);
    }

    // Start playback
    pVS1053->startPlayback();

    // Raise volume
    pVS1053->setVolume(vol);

    return true;
  }

  void stopPlayback(void) {

    // Stop the VS1053
    pVS1053->stopPlayback();

    delay(50);

    // Close the client
    client.stop();

    delay(50);
  }

  // Pause playback
  void pausePlayback(boolean state) {
    pVS1053->pausePlayback(state);
  }

  // Move the data to the VS1053
  void moveData(void) {

    int dataAvailable = client.available();
    if (dataAvailable > 0) {
      retryCount = 10;
      pVS1053->moveData(buffer, client.read(buffer, min(BUFFER_SIZE, dataAvailable)));
    } else {
      // Data seems to have stopped. This is either a temporary interruption
      // due to networking issues or the end of the song
      if (retryCount-- <= 0) {
        // End of song detected
        stopPlayback();
      } else {
        delay(50);
      }
    }
  }

  // Now playing screen
  void displaySongNowPlayingScreen(const char *albumName, const char *songName) {

    char line[40];

    processString(line, (char *)songName);

    tft.clearScreen();
    tft.drawRect(0, 0, tft.width(), tft.height(), COLOR_BLUE);

    tft.drawCenteredText(calcLineOffset(1), "- Now Playing -");
    tft.drawCenteredText(calcLineOffset(3), "Craig Lindley");
    tft.drawCenteredText(calcLineOffset(4), albumName);
    tft.setTextColor(COLOR_WHITE, COLOR_BLUE);
    tft.drawCenteredText(calcLineOffset(6), line);
    tft.setTextColor(COLOR_WHITE, COLOR_BLACK);

    // Display looping status
    displayLoopStatus(looping);
  }

  // Update the screen above with loop status
  void displayLoopStatus(bool loopSet) {
    if (loopSet) {
      tft.drawCenteredText(calcLineOffset(10), "Looping");
    } else {
      tft.drawCenteredText(calcLineOffset(10), "        ");
    }
  }

  // Perform any setup required for SD playback
  void webSetup() {
    // Scrape the website for album/song info
    ws.collectWebData();

    // Display the collected data
    // ws.displayCollectedData();

    // Initialize control variables
    looping = false;
    skipInput = false;

    // Set the initial state machne start state
    state = OP_POPULATE_LB;
  }

  // Realtime loop implementing FSM
  void webLoop() {

    // Run a Finite State Machine (FSM) to control the music player
    switch (state) {

      case OP_POPULATE_LB:
        {
          // Display the menu ListBox
          listBox.setDataSource(OPERATIONS_DS);
          listBox.clear();
          listBox.setCenterFlag(true);
          listBox.setTitle("- Operations -");
          listBox.addEntry("Sequential Play");
          listBox.addEntry("Random Play");

          listBox.doRepaint(true);

          // Next state
          state = OP_NAV_CHECK;
        }
        break;

      case OP_NAV_CHECK:
        {
          // Poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          if (result != Navigation::NAV_NONE) {
            updateTimeOut();
          }

          if (result == Navigation::NAV_UP) {
            listBox.selectionDown(true);
          }

          else if (result == Navigation::NAV_DOWN) {
            listBox.selectionUp(true);
          }

          else if (result == Navigation::NAV_BACK) {
            // Nothing to do here
          }

          else if (result == Navigation::NAV_SELECT) {
            // Next state
            state = OP_DISPATCH;
          }
        }
        break;

      case OP_DISPATCH:
        {
          switch (listBox.getSelectionIndex()) {
            case 0:
              // Sequential play selected
              playMode = SEQUENTIAL;

              // Next state
              state = AL_POPULATE_LB;
              break;

            case 1:
              // Random play selected
              playMode = RANDOM;

              // Next state
              state = AL_POPULATE_LB;
              break;
          }
        }
        break;

      case AL_POPULATE_LB:
        {
          listBox.setDataSource(ALBUMS_DS);
          listBox.clear();
          listBox.setCenterFlag(true);
          listBox.setSortFlag(true);
          listBox.setTitle("- Albums/CDs -");

          // Populate the list box with album names
          int numAlbums = ws.getAlbumCount();
          for (int i = 0; i < numAlbums; i++) {
            listBox.addEntry(ws.getAlbumName(i));
          }

          // Sort the album names
          listBox.prepareEntries();

          // Display the album names
          listBox.doRepaint(true);

          // Next state
          state = AL_NAV_CHECK;
        }
        break;

      case AL_NAV_CHECK:
        {
          // Poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          if (result != Navigation::NAV_NONE) {
            updateTimeOut();
          }

          if (result == Navigation::NAV_UP) {
            listBox.selectionDown(true);
          }

          else if (result == Navigation::NAV_DOWN) {
            listBox.selectionUp(true);
          }

          else if (result == Navigation::NAV_BACK) {
            // Back to operations menu
            listBox.setDataSource(OPERATIONS_DS);
            listBox.doRepaint(true);

            // Next state
            state = OP_NAV_CHECK;
          }

          else if (result == Navigation::NAV_SELECT) {
            // Next state
            state = SG_POPULATE_LB;
          }
        }
        break;

      case SG_POPULATE_LB:
        {
          currentAlbumIndex = 0;

          // Iterate through the albums names to find the correct index
          int numAlbums = ws.getAlbumCount();
          for (int i = 0; i < numAlbums; i++) {
            if (strcmp(listBox.getSelection(), ws.getAlbumName(i)) == 0) {
              currentAlbumIndex = i;
              break;
            }
          }

          listBox.setDataSource(SONGS_DS);
          listBox.clear();
          listBox.setCenterFlag(true);
          listBox.setSortFlag(true);
          listBox.setTitle("- Songs -");

          // Get count of songs
          int songCount = ws.getSongCount(currentAlbumIndex);

          for (int i = 0; i < songCount; i++) {
            listBox.addEntry(ws.getSongName(currentAlbumIndex, i));
          }

          // Sort the song names
          listBox.prepareEntries();

          // Display the song names
          listBox.doRepaint(true);

          // Next state
          state = SG_NAV_CHECK;
        }
        break;

      case SG_NAV_CHECK:
        {
          // Poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          if (result != Navigation::NAV_NONE) {
            updateTimeOut();
          }

          if (result == Navigation::NAV_UP) {
            listBox.selectionDown(true);
          }

          else if (result == Navigation::NAV_DOWN) {
            listBox.selectionUp(true);
          }

          else if (result == Navigation::NAV_BACK) {

            // Back to albums selection
            listBox.setDataSource(ALBUMS_DS);
            listBox.doRepaint(true);

            // Next state
            state = AL_NAV_CHECK;
          }

          else if (result == Navigation::NAV_SELECT) {
            // Song selected

            // Next state
            state = SG_PLAY;
          }
        }
        break;

      case SG_PLAY:
        {
          currentSongIndex = 0;

          // Iterate through the song names to find the correct index
          int numSongs = ws.getSongCount(currentAlbumIndex);
          for (int i = 0; i < numSongs; i++) {
            if (strcmp(listBox.getSelection(), ws.getSongName(currentAlbumIndex, i)) == 0) {
              currentSongIndex = i;
              break;
            }
          }

          // Show the song playing
          displaySongNowPlayingScreen(
            ws.getAlbumName(currentAlbumIndex), ws.getSongName(currentAlbumIndex, currentSongIndex));

          // Turn display back on if off for song change
          updateTimeOut();

          // Start playing the song
          String request =
            String("/MyCDs/") + ws.getAlbumName(currentAlbumIndex) + "/" + ws.getSongName(currentAlbumIndex, currentSongIndex);

          startPlayback(request.c_str());

          // Next state
          state = SG_SONGSTATUS_CHECK;
        }
        break;

      case SG_SONGSTATUS_CHECK:
        {
          // Has the song ended ?
          if (!isPlaying()) {
            // Song has ended
            // Are we looping on this song ?
            if (looping) {
              // Looping
            } else {
              // Not looping
              if (playMode == SEQUENTIAL) {
                listBox.selectionDown(false);
              } else {
                listBox.selectRandomEntry(false);
              }
            }
            // Next state
            state = SG_PLAY;
            break;
          }
          // Song is playing so poll the songs navigational rotary encoder
          byte result = songsNav.getNavigationDirection();

          // If skipInput is true we are trying to consume
          // the first button press because it should just
          // turn the backlight back on

          if (skipInput) {
            if (result != Navigation::NAV_NONE) {
              skipInput = false;
              updateTimeOut();

              // Stay in this state
              break;
            }
          }
          if (result != Navigation::NAV_NONE) {
            updateTimeOut();
          }

          if (result == Navigation::NAV_UP) {
            // Stop streaming
            stopPlayback();

            listBox.selectionDown(false);

            // Next state
            state = SG_PLAY;
          }

          else if (result == Navigation::NAV_DOWN) {
            // Stop streaming
            stopPlayback();

            listBox.selectionUp(false);

            // Next state
            state = SG_PLAY;
          }

          else if (result == Navigation::NAV_BACK) {
            stopPlayback();

            // Back to songs menu
            listBox.setDataSource(SONGS_DS);
            listBox.doRepaint(true);

            // Next state
            state = SG_NAV_CHECK;
          }

          else if (result == Navigation::NAV_SELECT) {
            // Select button during song playback toggles looping
            looping = !looping;

            displayLoopStatus(looping);
          }

          else if (millis() > displayTimeout) {
            // If skipInput is true
            // Ignore the button press that turns the display backlight on
            skipInput = true;
            tft.setBacklight(LOW);
          }
        }
        break;
    }
  }

protected:

  VS1053 *pVS1053;

  // Instantiate WiFi client
  WiFiClient client;

  int retryCount;

  // Buffer size
  static const int BUFFER_SIZE = VS1053_CHUNK_SIZE * 100;

  byte buffer[BUFFER_SIZE];

  // Instantiate web scraper
  WebScraper ws;

  uint8_t currentAlbumIndex;
  uint8_t currentSongIndex;
  boolean looping;
  boolean skipInput;

  enum PLAY_MODE { SEQUENTIAL,
                   RANDOM };
  enum PLAY_MODE playMode;

  // Finite State Machine states
  enum STATES {

    // Operation states
    OP_POPULATE_LB,
    OP_NAV_CHECK,
    OP_DISPATCH,

    // Album states
    AL_POPULATE_LB,
    AL_NAV_CHECK,

    // Song states
    SG_POPULATE_LB,
    SG_NAV_CHECK,
    SG_PLAY,
    SG_SONGSTATUS_CHECK,
  };

  // FSM state
  STATES state;
};

#endif
