#include <sys/_stdint.h>
/*
 * ESP32 Network Music Player - Web Scraper Code
 *
 * This code goes out to http://craigandheather.net/MyCDs
 * and collects all of my album names and album song filenames
 * to make them available for playback.
 *
 * Concept, Design and Implementation by: Craig A. Lindley
 * Last Update: 11/21/2023
 */

#ifndef WEBSCRAPER
#define WEBSCRAPER

#include <vector>
#include <string>
#include <HTTPClient.h>

class WebScraper {

public:

  // Return the count of albums found
  int getAlbumCount() {

    return albums.size();
  }

  // Get the album name at the index
  const char* getAlbumName(int albumIndex) {

    return albums[albumIndex]->albumName.c_str();
  }

  // Return the count of MP3 songs in album specified by index
  int getSongCount(int albumIndex) {
    int count = 0;

    // Get total songs in specified album
    uint16_t totalSongs = albums[albumIndex]->songNames.size();
    for (int j = 0; j < totalSongs; ++j) {
      const char* songName = albums[albumIndex]->songNames[j].c_str();
      bool isMP3 = strstr(songName, ".mp3") != NULL;
      if (isMP3) {
        count++;
      }
    }
    return count;
  }

  // Return the count of MP3 songs in album specified by index
  const char* getSongName(int albumIndex, int songIndex) {

    return albums[albumIndex]->songNames[songIndex].c_str();
  }

  // Return the URL for the specified album song
  const char* getSongURL(int albumIndex, int songIndex) {

    return (albumsURL + getAlbumName(albumIndex) + "/" + getSongName(albumIndex, songIndex)).c_str();
  }

  // Scrap the web page for my album/song data
  void collectWebData() {

    // Attempt to get album names from website
    String payload = doGet(albumsURL.c_str());

    if (payload.length() != 0) {
      // GET successful so parse the hrefs for the names
      parseAlbumNameHRefs(payload);
    }

    // Now for each album, gather the song filenames
    for (int index = 0; index < albums.size(); ++index) {
      ALBUM_SONGS* as = albums[index];
      String songsURL = albumsURL + as->albumName.c_str() + "/";

      // Attempt to get song filename from website
      payload = doGet(songsURL.c_str());

      if (payload.length() != 0) {
        // GET successful so parse the hrefs for the names
        parseSongNameHRefs(payload, index);
      }
    }
  }

  // Display the collected data
  void displayCollectedData() {

    for (int i = 0; i < albums.size(); ++i) {
      Serial.printf("Album name: %s\n", albums[i]->albumName.c_str());

      for (int j = 0; j < albums[i]->songNames.size(); j++) {
        const char* songName = albums[i]->songNames[j].c_str();
        Serial.printf("\tSong: %s\n", songName);
      }
    }
  }

protected:

  // Protected data
  // Path to my online albums
  String albumsURL = "http://craigandheather.net/MyCDs/";

  typedef struct {
    std::string albumName;
    std::vector<std::string> songNames;
  } ALBUM_SONGS;

  std::vector<ALBUM_SONGS*> albums;

  // Protected methods

  // Parse out the albums names from the HTML
  void parseAlbumNameHRefs(String s) {
    int startIndex = 0;
    int endIndex = 0;
    bool done = false;

    // Skip extraneous data
    startIndex = s.indexOf("Parent Directory");
    if (startIndex != 1) {

      while (!done) {
        startIndex = s.indexOf("href=\"", startIndex);
        if (startIndex == -1) {
          done = true;
          continue;
        }
        endIndex = s.indexOf("/\"", startIndex + 6);

        // Create new ALBUM_SONG structure
        ALBUM_SONGS* as = new ALBUM_SONGS;

        // Clear the song names vector
        as->songNames.clear();

        // Log the album name
        as->albumName = s.substring(startIndex + 6, endIndex).c_str();
        albums.push_back(as);
        startIndex = endIndex;
      }
    }
  }

  // Parse out the song names from the HTML
  void parseSongNameHRefs(String s, int index) {

    int startIndex = 0;
    int endIndex = 0;
    bool done = false;

    // Skip extraneous data
    startIndex = s.indexOf("Parent Directory");
    if (startIndex != -1) {

      while (!done) {
        startIndex = s.indexOf("href=", startIndex);
        if (startIndex == -1) {
          done = true;
          continue;
        }
        endIndex = s.indexOf("\">", startIndex + 6);

        // Add song name to songNames vector
        String songName = s.substring(startIndex + 6, endIndex);
        if (songName.endsWith(".mp3")) {
          //          Serial.println(songName);
          albums[index]->songNames.push_back(songName.c_str());
        }
        startIndex = endIndex;
      }
    }
  }

  // Do a GET to the specified URL and return the returned payload
  String doGet(const char* url) {

    String payload = "";

    //Check WiFi connection status
    if (WiFi.status() == WL_CONNECTED) {
      HTTPClient http;

      // Path to URL
      http.begin(url);

      // Send HTTP GET request
      int httpResponseCode = http.GET();

      if (httpResponseCode > 0) {
        payload = http.getString();
      } else {
        Serial.print("Error code: ");
        Serial.println(httpResponseCode);
      }
      // Free resources
      http.end();
    } else {
      Serial.println("WiFi Disconnected");
    }
    return payload;
  }
};

#endif
